

/******************************************************************************
*
* Copyright (C) 2002 Hugo PEREIRA <mailto: hugo.pereira@free.fr>
*
* This is free software; you can redistribute it and/or modify it under the
* terms of the GNU General Public License as published by the Free Software
* Foundation; either version 2 of the License, or (at your option) any later
* version.
*
* This software is distributed in the hope that it will be useful, but WITHOUT
* Any WARRANTY; without even the implied warranty of MERCHANTABILITY or
* FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
* for more details.
*
* You should have received a copy of the GNU General Public License along with
* this program.  If not, see <http://www.gnu.org/licenses/>.
*
*******************************************************************************/

#include "RevisionFrame.h"
#include "RevisionFrame.moc"

#include "AnimatedTreeView.h"
#include "Command.h"
#include "ContextMenu.h"
#include "CvsInterface.h"
#include "IconNames.h"
#include "IconEngine.h"
#include "InformationDialog.h"
#include "RevisionDelegate.h"

#include <QHeaderView>
#include <QItemSelectionModel>
#include <QLayout>

//________________________________________________________________
RevisionFrame::RevisionFrame( QWidget* parent ):
    QWidget( parent ),
    Counter( "RevisionFrame" ),
    defaultWidth_( -1 ),
    interface_( new CvsInterface( this ) ),
    visibilityOptionName_( "REVISION_FRAME_VISIBILITY" )
{

    Debug::Throw( "RevisionFrame::RevisionFrame.\n" );
    QVBoxLayout* layout = new QVBoxLayout();
    layout->setMargin(0);
    layout->setSpacing(5);
    setLayout( layout );

    _installActions();

    //! list
    layout->addWidget( list_ = new AnimatedTreeView( this ), 1 );
    list_->setModel( &model_ );
    list_->setSortingEnabled( false );

    // replace item delegate
    if( list_->itemDelegate() ) list_->itemDelegate()->deleteLater();
    list_->setItemDelegate( delegate_ = new RevisionDelegate( list_ ) );

    list_->setMask( 1 );
    list_->forceAlternatingRowColors();
    list_->header()->hide();
    list_->setEnabled( false );

    QMenu* menu( new ContextMenu( list_ ) );
    menu->addAction( &editLocalAction() );
    menu->addAction( &editRevisionAction() );
    menu->addAction( &diffAction() );

    connect( this, SIGNAL(itemResized(QModelIndex)), delegate_, SIGNAL(sizeHintChanged(QModelIndex)) );

    // connections to CvsInterface
    connect( list_, SIGNAL(activated(QModelIndex)), SLOT(_editRevision()) );
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SLOT(_updateActions()) );
    connect( list_->selectionModel(), SIGNAL(selectionChanged(QItemSelection,QItemSelection)), SIGNAL(selectionChanged()) );
    connect( list_->selectionModel(), SIGNAL(currentChanged(QModelIndex,QModelIndex)), SLOT(_currentRevisionChanged(QModelIndex)) );
    connect( interface_, SIGNAL(revisionsAvailable(File,Revision::List)), SLOT(_showRevisions(File,Revision::List)) );
    connect( interface_, SIGNAL(commandCompleted(CvsCommand)), SLOT(commandCompleted(CvsCommand)) );
    _updateActions();

    // visibility
    if( XmlOptions::get().contains( visibilityOptionName_ ) )
    { visibilityAction().setChecked( XmlOptions::get().get<bool>( visibilityOptionName_ ) ); }

    return;
}

//________________________________________________________________
void RevisionFrame::clear( void )
{

    Debug::Throw( "RevisionFrame::clear.\n" );
    list_->initializeAnimation();
    model_.clear();
    file_.clear();

    list_->startAnimation();
}

//________________________________________________________________
void RevisionFrame::setFile( const File& file )
{
    Debug::Throw( "RevisionFrame::setFile.\n" );

    // see if file has changed
    if( file == file_ ) return;

    // reload entries
    if( !file.isEmpty() && (file_.isEmpty() || file.path() != file_.path() ) && interface_->isCvsPath( file.path() ) )
    {

        interface_->updateEntries( file.path() );

    }

    file_ = file;
    if( file.isEmpty() )
    {
        clear();
        list_->setEnabled( false );
        return;
    }

    // clear list
    list_->initializeAnimation();
    model_.clear();
    list_->startAnimation();

    // try load revisions from cache
    // then from CVS
    if( interface_->isCvsFile( file ) )
    {

        _showRevisionsFromLocalCache();

        // delayed CVS reload
        timer_.start( 500, this );

    } else list_->setEnabled( false );

}

//_______________________________________________________________________
void RevisionFrame::commandCompleted( const CvsCommand& command )
{

    Debug::Throw( "RevisionFrame::_commandCompleted.\n" );
    switch( command.type() )
    {
        case CvsCommand::Commit:
        _reload();
        break;

        case CvsCommand::Log:
        if( !localCache_.empty() )
        {
            std::sort( localCache_.begin(), localCache_.end() );
            localCache_.erase( std::unique( localCache_.begin(), localCache_.end() ), localCache_.end() );
            cache_[file()] = localCache_;
        }
        break;

        default:
        break;
    }
}

//__________________________________________________________________________
void RevisionFrame::_updateActions( void )
{

    Debug::Throw( "RevisionFrame::_updateActions.\n" );

    bool fileExists( file().exists() );
    bool hasSelection( !list_->selectionModel()->selectedRows().empty() );

    editLocalAction().setEnabled( fileExists );
    editRevisionAction().setEnabled( hasSelection );
    diffAction().setEnabled( hasSelection && fileExists );

}

//__________________________________________________________________________
void RevisionFrame::timerEvent( QTimerEvent* event )
{
    if( event->timerId() == timer_.timerId() )
    {

        timer_.stop();
        _reload();

    } else if( event->timerId() == resizeTimer_.timerId() ) {

        resizeTimer_.stop();
        for( int row = 0; row < model_.rowCount(); ++row )
        { emit itemResized( model_.index( row, 0 ) ); }

        // make sure current index is visible
        if( list_->selectionModel()->hasSelection() )
        { list_->scrollTo( list_->selectionModel()->selectedIndexes().front() ); }

    } else return QWidget::timerEvent( event );

}

//__________________________________________________________________________
void RevisionFrame::resizeEvent( QResizeEvent* event )
{

    // check if width has changed
    if( event->oldSize().width() != event->size().width() )
    { resizeTimer_.start( 200, this ); }

    QWidget::resizeEvent( event );

}

//__________________________________________________________________________
void RevisionFrame::_reload( void )
{

    Debug::Throw( "RevisionFrame::_reload.\n" );
    if( interface_->isCvsPath( file().path() ) ) interface_->updateEntries( file().path() );
    if( interface_->isCvsFile( file() ) )
    {
        interface_->abort();
        interface_->logRevisions( file() );
    }

}


//______________________________________________________________________
void RevisionFrame::_editLocal( void )
{

    Debug::Throw( "RevisionFrame::_editLocal.\n" );
    ( Command( XmlOptions::get().raw( "EDITOR" ) ) << file().expand() ).run();

}

//______________________________________________________________________
void RevisionFrame::_editRevision( void )
{

    Debug::Throw( "RevisionFrame::_editRevision.\n" );

    // check selection, retrieve revision
    QModelIndex current( list_->currentIndex() );
    if( !current.isValid() )
    {
        InformationDialog( this, tr( "No revision selected. <Edit> canceled." ) ).exec();
        return;
    }

    disconnect( interface_, SIGNAL(revisionAvailable(File,File)), this, SLOT(_diff(File,File)) );
    disconnect( interface_, SIGNAL(revisionAvailable(File,File)), this, SLOT(_editRevision(File,File)) );
    connect( interface_, SIGNAL(revisionAvailable(File,File)), this, SLOT(_editRevision(File,File)) );
    interface_->checkOutRevision( file(), model_.get( current ).name() );

}

//_______________________________________________________________________
void RevisionFrame::_diff( void )
{

    Debug::Throw( "RevisionFrame::_diff.\n" );

    // check selection, retrieve revision
    QModelIndex current( list_->currentIndex() );
    if( !current.isValid() )
    {
        InformationDialog( this, tr( "No revision selected. <Diff> canceled." ) ).exec();
        return;
    }

    disconnect( interface_, SIGNAL(revisionAvailable(File,File)), this, SLOT(_editRevision(File,File)) );
    disconnect( interface_, SIGNAL(revisionAvailable(File,File)), this, SLOT(_diff(File,File)) );
    connect( interface_, SIGNAL(revisionAvailable(File,File)), this, SLOT(_diff(File,File)) );
    interface_->checkOutRevision( file(), model_.get( current ).name() );

}

//__________________________________________________________________________
void RevisionFrame::_showRevisions( File file, Revision::List revisionList )
{
    Debug::Throw( "RevisionFrame::_showRevisions.\n" );

    // do nothing if file has changed
    if( file != file_ ) return;

    list_->setEnabled( !revisionList.isEmpty() );

    // add new items
    for( Revision::List::const_iterator iter = revisionList.begin(); iter != revisionList.end(); ++iter )
    { localCache_ << *iter; }
    model_.add( revisionList );

    // scroll to last item or selected
    if( list_->currentIndex().isValid() ) list_->scrollTo( list_->currentIndex() );
    else if( model_.rowCount() ) list_->scrollTo( model_.index( model_.rowCount()-1, 0 ) );

}

//______________________________________________________________________
void RevisionFrame::_editRevision( File, File destination )
{

    Debug::Throw() << "RevisionFrame::_editRevision - file: " << destination << endl;
    ( Command( XmlOptions::get().raw( "EDITOR" ) ) << destination.expand() ).run();
    return;

}

//_______________________________________________________________________
void RevisionFrame::_diff( File source, File destination )
{

    Debug::Throw() << "RevisionFrame::_diff - file: " << destination << endl;
    ( Command( XmlOptions::get().raw( "DIFF_EDITOR" ) ) << source.expand() << destination.expand() ).run();
    return;

}

//__________________________________________________________________________
void RevisionFrame::_installActions( void )
{

    Debug::Throw( "RevisionFrame::_installActions.\n" );

    // toggle visibility action
    addAction( visibilityAction_ = new QAction( tr( "Show Revision List" ), this ) );
    visibilityAction_->setCheckable( true );
    visibilityAction_->setChecked( true );
    connect( visibilityAction_, SIGNAL(toggled(bool)), SLOT(_toggleVisibility(bool)) );

    // edit local file action
    addAction( editLocalAction_ = new QAction( IconEngine::get( IconNames::Edit ), tr( "Edit Local File" ), this ) );
    connect( editLocalAction_, SIGNAL(triggered()), SLOT(_editLocal()) );

    // edit revision action
    addAction( editRevisionAction_ = new QAction( IconEngine::get( IconNames::Edit ), tr( "Edit Revision" ), this ) );
    connect( editRevisionAction_, SIGNAL(triggered()), SLOT(_editRevision()) );

    // diff local and revision
    addAction( diffAction_ = new QAction( tr( "Diff Local and Revision" ), this ) );
    connect( diffAction_, SIGNAL(triggered()), SLOT(_diff()) );

}

//__________________________________________________________________________
void RevisionFrame::_showRevisionsFromLocalCache( void )
{
    Debug::Throw( "RevisionFrame::_showRevisionsFromLocalCache.\n" );

    // retrieve local_cache
    localCache_ = cache_[ file() ];
    if( localCache_.empty() )
    {
        list_->setEnabled( false );
        return;
    }

    // add new items
    for( Revision::List::const_iterator iter = localCache_.begin(); iter != localCache_.end(); ++iter )
    { model_.add( *iter ); }

    // scroll to last item or selected
    if( list_->currentIndex().isValid() ) list_->scrollTo( list_->currentIndex() );
    else if( model_.rowCount() ) list_->scrollTo( model_.index( model_.rowCount()-1, 0 ) );

    list_->setEnabled( true );

}
